Proximo = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceComm-2.0", "AceHook-2.1")
local L = AceLibrary("AceLocale-2.2"):new("Proximo")
local BZ = AceLibrary("Babble-Zone-2.2")
local BC = AceLibrary("Babble-Class-2.2")

Proximo.title = "Proximo"
Proximo.OnCommReceive = {}

local prox_list = {}
local unit = 1
local arena_start = false
local mainassist = "target"

local UnitName = UnitName
local UnitClass = UnitClass
local UnitHealth = UnitHealth
local UnitExists = UnitExists
local UnitCanAttack = UnitCanAttack
local UnitIsPlayer = UnitIsPlayer
local InCombatLockdown = InCombatLockdown
local GetNumPartyMembers = GetNumPartyMembers
local pairs = pairs
local GetPlayerBuff = GetPlayerBuff
local GetPlayerBuffTexture = GetPlayerBuffTexture
local UNKNOWNOBJECT = UNKNOWNOBJECT

----------------------
--table caching
local new, del
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return { ... }
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end

local function has_arena_started()
	if not arena_start then
		local pindex, icon
		for i = 1, 16 do
			pindex = GetPlayerBuff(i, "HELPFUL")
			icon = GetPlayerBuffTexture(pindex)
			if icon and icon:match("Spell_Nature_WispSplode") then
				arena_start = false
				return
			end
		end
		arena_start = true
	end
end

local function valid_unit(unitid)
	if arena_start and UnitExists(unitid) and UnitName(unitid) and UnitIsPlayer(unitid) and UnitCanAttack("player",unitid) and not UnitIsCharmed(unitid) then
		return true
	else
		return false
	end 
end

----------------------
--On start
function Proximo:OnInitialize()
	self:SetupOptions()
	self:SetDefaultCommPriority("ALERT")
	self:SetCommPrefix(self.title)
	self:RegisterMemoizations {BC["Druid"],BC["Hunter"],BC["Mage"],BC["Paladin"],BC["Priest"],BC["Rogue"],BC["Shaman"],BC["Warlock"],BC["Warrior"]}
  self:RegisterDB("ProximoDB")
	self:RegisterDefaults("profile", self:GetDefaultConfig())   
	self:RegisterChatCommand({"/proximo", "/prox"}, self.options)
end

----------------------
--Disabled
function Proximo:OnDisable()
	self:ResetList()
	if self.frame then
		self.frame:Hide()
	end
end

----------------------
--Loaded
function Proximo:OnEnable()
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterComm(self.title, "GROUP")
end

----------------------
--Player Zoned
function Proximo:PLAYER_ENTERING_WORLD()
	self:ResetList()
	arena_start = false
	self:ScheduleEvent(self.ZoneCheck, 5, self)
end

----------------------
--Hostile Death
function Proximo:CHAT_MSG_COMBAT_HOSTILE_DEATH(msg)
  local k, v
	for k, v in pairs(prox_list) do
		if msg == string.format(UNITDIESOTHER, k) then
			v.health = 0
			break
		end
  end
end

----------------------
--Zone Check
function Proximo:ZoneCheck()
	local zone = GetMinimapZoneText()
	if zone == BZ["Ruins of Lordaeron"] or zone == BZ["Nagrand Arena"] or zone == BZ["Blade's Edge Arena"] then
		if self.frame==nil then
			self:MakeFrame()
		end
		self.frame:Show()
		self:RegisterEvent("UNIT_HEALTH")
 		self:RegisterEvent("CHAT_MSG_COMBAT_HOSTILE_DEATH")
 		self:RegisterEvent("UPDATE_MOUSEOVER_UNIT") 
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:ScheduleRepeatingEvent("proximorefresh", self.FrameRefresh, .1, self)
		self:ScheduleRepeatingEvent("proximoparty", self.PartyTargets, 1, self)
		self:ScheduleRepeatingEvent("proximosync", self.Sync, 1, self)
		--enable nameplate tracking if they are on
		if NAMEPLATES_ON then
			self:CreateUnitUpdate()
		end
	else 
		if self:IsEventRegistered("UNIT_HEALTH") then self:UnregisterEvent("UNIT_HEALTH") end
	  if self:IsEventRegistered("CHAT_MSG_COMBAT_HOSTILE_DEATH") then self:UnregisterEvent("CHAT_MSG_COMBAT_HOSTILE_DEATH") end
	  if self:IsEventRegistered("UPDATE_MOUSEOVER_UNIT") then self:UnregisterEvent("UPDATE_MOUSEOVER_UNIT") end
	 	if self:IsEventRegistered("PLAYER_TARGET_CHANGED") then self:UnregisterEvent("PLAYER_TARGET_CHANGED") end
	  if self:IsEventScheduled("proximorefresh") then self:CancelScheduledEvent("proximorefresh") end
	  if self:IsEventScheduled("proximosync") then self:CancelScheduledEvent("proximosync") end
	  if self:IsEventScheduled("proximoparty") then self:CancelScheduledEvent("proximoparty") end
	  if NAMEPLATES_ON then
			self:StopUnitUpdate()
		end
		if self.frame then self.frame:Hide() end
	end
end


----------------------
--Add a unit to local list
function Proximo:AddToList(name, class, health)
	if not name or name == UNKNOWNOBJECT then return end
	local t = new()
	t.class = class
	t.health = health
	t.click = false
	t.unit = unit
  prox_list[name] = t
  --setup frame class
  self:FrameClass(self.frame.buttons[unit],class)
  --update unit count
  unit = unit + 1
  if unit > 5 then unit = 5 end
  --sent annoucement
  self:SendAnnouncement(name, class)
end

----------------------
--Reset local list
function Proximo:ResetList()
	local k, v
	for k, v in pairs(prox_list) do
		prox_list[k] = del(v)
	end
	unit = 1
	self:FrameClean()
end

----------------------
--get local list
function Proximo:GetList()
	return prox_list
end

----------------------
--Check unit for updates
function Proximo:CheckUnitIDChanges(unitid)
	if valid_unit(unitid) then
		local name, health=UnitName(unitid), UnitHealth(unitid)
		if not prox_list[name] then 
	 		local class = UnitClass(unitid)
			self:AddToList(name, class, health)
		else
			prox_list[name].health = health
		end
	end
end

----------------------
--Mouse over event
function Proximo:UPDATE_MOUSEOVER_UNIT()
	self:CheckUnitIDChanges("mouseover")
end

----------------------
--Target changed event
function Proximo:PLAYER_TARGET_CHANGED()
	self:CheckUnitIDChanges("target")
end

----------------------
--Unit Health event
function Proximo:UNIT_HEALTH(arg1)
 if arg1=="target" or arg1=="focus" then
	 self:CheckUnitIDChanges(arg1)
 end
end

----------------------
--Scan party for target updates
function Proximo:PartyTargets()
	local i, party
	for i=1, GetNumPartyMembers() do
		party = "party"..i.."target"
		self:CheckUnitIDChanges(party)
	end
	--update arena_start flag, no need to update once combat starts
	if not InCombatLockdown() then
		has_arena_started()
	end
	--update MA while arena has not started
	if not arena_start then
		self:SetMainAssist(self.db.profile.ma)
	end
end

----------------------
--Set the MA based on a name
function Proximo:SetMainAssist(ma)
	if UnitName("player") == ma then 
		mainassist = "target" 
	else
		local i, party
		for i=1, GetNumPartyMembers() do
			party = "party"..i
			if UnitName(party) == ma then
				mainassist = party.."target" 
				return
			end
		end
	end
end

----------------------
--Sync all data with clients
function Proximo:Sync()
	if arena_start and self.db.profile.sync then
		local k, v
		for k, v in pairs(prox_list) do
			self:SendCommMessage("GROUP", "ReceiveSync", k, v.class, v.health)
		end
	end
end

----------------------
--Receive a sync
function Proximo.OnCommReceive:ReceiveSync(prefix, sender, distribution, name, class, health)
	--if arena has started
	if arena_start and self.db.profile.sync then
		--add if new name
		if not prox_list[name] then
	 		self:AddToList(name, class, health)
	 	--don't sync health for your target or focus (you will always have best data)
	 	elseif UnitName("target") ~= name and UnitName("focus") ~= name then
	 		--only update if you don't see them as dead, wait for client to see them as alive
	 		if prox_list[name].health ~= 0 then
	  		prox_list[name].health = health
	  	end
		end
	end
end

----------------------
--Receive a MA
function Proximo.OnCommReceive:ReceiveMainAssist(prefix, sender, distribution, ma)
	self.db.profile.ma = ma
	self:SetMainAssist(ma)
	self:Print(L["MainAssist"]..": "..ma)
end

----------------------
--Refresh the frame display
function Proximo:FrameRefresh() 
	local k, v, name, class, health, unit, colortext, alpha
	--update each frame
	for k, v in pairs(prox_list) do
    name, class, health, unit = k, v.class, v.health, v.unit
		alpha, colortext = 0, "|cFFFFFFFF"
		--set colors
		if health <= 0 then
			colortext = "|cFF444444"
		elseif health <= 33 then
			colortext = "|cFFFF0000"
		elseif health <= 66 then
			colortext = "|cFFFFFF00"
		end
    --update status bar
		self.frame.buttons[unit].healthBar:SetValue(health)
		self.frame.buttons[unit].healthtext:SetText(colortext..health.."%|r")
		--update highlight
		if UnitName(mainassist) == name then alpha = 1 end
		self.frame.buttons[unit].healthBar.highlight:SetBackdropBorderColor(1,1,0,alpha)
		--update macros if not updated already and not in combat
		if not v.click and not InCombatLockdown() then 
			self.frame.buttons[unit]:SetAttribute("macrotext1", "/target "..name)
			if self.db.profile.focus then
				self.frame.buttons[unit]:SetAttribute("macrotext2", "/target "..name.."\n/focus\n/targetlasttarget")
			end
			v.click = 1
		end
		--Update unit text
    if v.click then 
			self.frame.buttons[unit].text:SetText(colortext..name.."|r")
		else
			self.frame.buttons[unit].text:SetText(colortext.."** "..name.." **|r")
    end
  end
end